# -*- Mode: perl; indent-tabs-mode: nil -*-
#
# The contents of this file are subject to the Mozilla Public
# License Version 1.1 (the "License"); you may not use this file
# except in compliance with the License. You may obtain a copy of
# the License at http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS
# IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
# implied. See the License for the specific language governing
# rights and limitations under the License.
#
# The Original Code is the Bugzilla Bug Tracking System.
#
# Contributor(s): Marc Schumann <wurblzap@gmail.com>
#                 Max Kanat-Alexander <mkanat@bugzilla.org>
#                 Mads Bondo Dydensborg <mbd@dbc.dk>
#                 Tsahi Asher <tsahi_75@yahoo.com>
#                 Elliotte Martin <emartin@everythingsolved.com>
#                 Bradley Baetz <bbaetz@acm.org>


package Bugzilla::WebService::Bug;

use strict;
use base qw(Bugzilla::WebService);
import SOAP::Data qw(type);

use Bugzilla::Constants;
use Bugzilla::Error;
use Bugzilla::Field;
use Bugzilla::WebService::Constants;
use Bugzilla::Bug;
use Bugzilla::BugMail;
use Bugzilla::Search;
use Bugzilla::Util;

#############
# Constants #
#############

# This maps the names of internal Bugzilla bug fields to things that would
# make sense to somebody who's not intimately familiar with the inner workings
# of Bugzilla. (These are the field names that the WebService uses.)
use constant FIELD_MAP => {
    status      => 'bug_status',
    severity    => 'bug_severity',
    description => 'comment',
    summary     => 'short_desc',
    platform    => 'rep_platform',
};

use constant GLOBAL_SELECT_FIELDS => qw(
    bug_severity
    bug_status
    op_sys
    priority
    rep_platform
    resolution
);

use constant PRODUCT_SPECIFIC_FIELDS => qw(version target_milestone component);

######################################################
# Add aliases here for old method name compatibility #
######################################################

BEGIN { *get_bugs = \&get }

###########
# Methods #
###########

sub get {
    my ($self, $params) = @_;
    my $ids = $params->{ids};
    my $products = $params->{products};
    (defined $ids || defined $products)
        || ThrowCodeError('param_required', { param => 'ids or products' });

    my $limit = $params->{limit};
    if ($limit) {
        detaint_natural($limit)
            || ThrowUserError('invalid_maxrows', { maxrows => $params->{limit} });
    }
    my $offset = $params->{offset};
    if(defined($offset) && !$limit) {
        ThrowCodeError('param_required', { param => 'limit' });
    }
    elsif (defined $offset) {
        detaint_natural($offset)
            || ThrowUserError('invalid_offset', { offset => $params->{offset} });
    }

    # Use Search.pm so that all the constraints can be applied simply
    # including visibility limits. All parameter validation will happen
    # there.

    # Build up a boolean chart
    # Search takes an object that has a ->param method; CGI suffices.
    my $q = new CGI({});

    # First, bug and product constraints
    # These are a union
    my $chart = 0;
    my $col = 0;
    if (defined $ids) {
        # Need to handle bug ids and aliases
        # We *do* validate these, so that any explicitly
        # listed entries throw an error if they're not viewable
        my @ids;
        foreach my $bug_id (@$ids) {
            ValidateBugID($bug_id);
            push @ids, $bug_id;
        }
        $q->param("field$chart-0-$col", 'bug_id');
        $q->param("type$chart-0-$col", 'anyexact');
        $q->param("value$chart-0-$col", join(',', @ids));
        $col++;
    }
    if (defined $products) {
        # Can't use anyexact in case the product contains a comma
        foreach my $prod (@$products) {
            $q->param("field$chart-0-$col", 'product');
            $q->param("type$chart-0-$col", 'equals');
            $q->param("value$chart-0-$col", $prod);
            $col++;
        }
    }
    $chart++;

    # Now do changed_since
    my $changed_since = $params->{changed_since};
    if (defined $changed_since) {
        # Format incoming iso8601 dateTime correctly
        # This will error out if the incoming value isn't parseable
        # XXX - shouldn't Bugzilla::Search be doing this for date
        # searches?
        $changed_since = Bugzilla::Search::SqlifyDate($changed_since);

        # No greaterthanorequal, so do NOT lessthan
        $q->param("negate$chart", 1);
        $q->param("field$chart-0-0", 'delta_ts');
        $q->param("type$chart-0-0", 'lessthan');
        $q->param("value$chart-0-0", $changed_since);
        $chart++;
    }

    my $s = new Bugzilla::Search('fields' => ['bugs.bug_id'],
                                 'params' => $q,
                                 'order' => ['bugs.bug_id']);

    my $dbh = Bugzilla->dbh;

    my $sql = $s->getSQL();
    $sql .= ' ' . $dbh->sql_limit($limit, $offset) if ($limit);

    my $bugs = $dbh->selectcol_arrayref($sql);

    my @return;
    foreach my $bug_id (@$bugs) {
        # Note - checking if the user can see the bug is handled by Search.pm
        my $bug = new Bugzilla::Bug($bug_id);

        # Timetracking fields are deleted if the user doesn't belong to
        # the corresponding group.
        unless (Bugzilla->user->in_group(Bugzilla->params->{'timetrackinggroup'})) {
            delete $bug->{'estimated_time'};
            delete $bug->{'remaining_time'};
            delete $bug->{'deadline'};
        }
        # This is done in this fashion in order to produce a stable API.
        # The internals of Bugzilla::Bug are not stable enough to just
        # return them directly.
        my $creation_ts = $self->datetime_format($bug->creation_ts);
        my $delta_ts    = $self->datetime_format($bug->delta_ts);
        my $status      = $bug->status;
        my %item;
        $item{'id'}               = type('int')->value($bug->bug_id);

        $item{'assigned_to'}      = type('string')->value($bug->assigned_to->login);
        $item{'component'}        = type('string')->value($bug->component);
        $item{'creation_time'}    = type('dateTime')->value($creation_ts);
        $item{'dupe_of'}          = type('int')->value($bug->dup_id)
            if $bug->dup_id;
        $item{'is_open'}          = type('boolean')->value($status->is_open);
        $item{'last_change_time'} = type('dateTime')->value($delta_ts);
        $item{'priority'}         = type('string')->value($bug->priority);
        $item{'product'}          = type('string')->value($bug->product);
        $item{'resolution'}       = type('string')->value($bug->resolution);
        $item{'severity'}         = type('string')->value($bug->bug_severity);
        $item{'status'}           = type('string')->value($status->name);
        $item{'summary'}          = type('string')->value($bug->short_desc);

        # Let people poke inside if they know what they're doing...
        $item{'internals'}        = $bug;

        if (Bugzilla->params->{'usebugaliases'}) {
            $item{'alias'} = type('string')->value($bug->alias);
        }
        else {
            # For API reasons, we always want the value to appear, we just
            # don't want it to have a value if aliases are turned off.
            $item{'alias'} = undef;
        }

        push(@return, \%item);
    }

    return { bugs => \@return };
}


sub create {
    my ($self, $params) = @_;

    Bugzilla->login(LOGIN_REQUIRED);

    my %field_values;
    foreach my $field (keys %$params) {
        my $field_name = FIELD_MAP->{$field} || $field;
        $field_values{$field_name} = $params->{$field}; 
    }

    # WebService users can't set the creation date of a bug.
    delete $field_values{'creation_ts'};

    my $bug = Bugzilla::Bug->create(\%field_values);

    Bugzilla::BugMail::Send($bug->bug_id, { changer => $bug->reporter->login });

    return { id => type('int')->value($bug->bug_id) };
}

sub legal_values {
    my ($self, $params) = @_;
    my $field = FIELD_MAP->{$params->{field}} || $params->{field};

    my @custom_select =
        Bugzilla->get_fields({ type => FIELD_TYPE_SINGLE_SELECT });
    # We only want field names.
    @custom_select = map {$_->name} @custom_select;

    my $values;
    if (grep($_ eq $field, GLOBAL_SELECT_FIELDS, @custom_select)) {
        $values = get_legal_field_values($field);
    }
    elsif (grep($_ eq $field, PRODUCT_SPECIFIC_FIELDS)) {
        my $id = $params->{product_id};
        defined $id || ThrowCodeError('param_required',
            { function => 'Bug.legal_values', param => 'product_id' });
        grep($_->id eq $id, @{Bugzilla->user->get_accessible_products})
            || ThrowUserError('product_access_denied', { product => $id });

        my $product = new Bugzilla::Product($id);
        my @objects;
        if ($field eq 'version') {
            @objects = @{$product->versions};
        }
        elsif ($field eq 'target_milestone') {
            @objects = @{$product->milestones};
        }
        elsif ($field eq 'component') {
            @objects = @{$product->components};
        }

        $values = [map { $_->name } @objects];
    }
    else {
        ThrowCodeError('invalid_field_name', { field => $params->{field} });
    }

    my @result;
    foreach my $val (@$values) {
        push(@result, type('string')->value($val));
    }

    return { values => \@result };
}

sub add_comment {
    my ($self, $params) = @_;
    
    #The user must login in order add a comment
    Bugzilla->login(LOGIN_REQUIRED);
    
    # Check parameters
    defined $params->{id} 
        || ThrowCodeError('param_required', { param => 'id' });
    ValidateBugID($params->{id});
    
    my $comment = $params->{comment}; 
    defined $comment
        || ThrowCodeError('param_required', { param => 'comment' });
    
    my $bug = new Bugzilla::Bug($params->{id});
    
    Bugzilla->user->can_edit_product($bug->product_id)
        || ThrowUserError("product_edit_denied", {product => $bug->product});
        
    # Append comment
    $bug->add_comment($comment, { isprivate => $params->{private},
                                  work_time => $params->{work_time} });
    $bug->update();
    
    # Send mail.
    Bugzilla::BugMail::Send($bug->bug_id, { changer => Bugzilla->user->login });
    
    my $dbh = Bugzilla->dbh;
    my $comment_id = $dbh->bz_last_key('longdescs', 'comment_id');
    return {'comment_id' => type('int')->value($comment_id) };
}

sub comments {
    my ($self, $params) = @_;

    my $bug_ids = $params->{bug_ids};
    my $comment_ids = $params->{comment_ids};
    (defined $bug_ids || defined $comment_ids)
        || ThrowCodeError('param_required', { param => 'bug_ids or comment_ids' });

    # This is a bit ugly. We can't just get the comments matching
    # 'comment_ids' because the only way to get the comment number
    # is to get all the comments.
    # Also, we need to record all comments from passed in bugs,
    # but only the requested comments from things not passed in,
    # so we need to keep track of the difference.
    # We also need to check for validity up front. So what we do is:
    #  - Get the comments
    #  - Check for validity
    #  - Have a hash that keeps track of whether we want all comments
    #    or not.
    # This interface is sucky!
    # There are possible perf issues, but it *should* be OK for now,
    # given the expected use cases
    $bug_ids = { map { $_ => 1 } @$bug_ids };
    $comment_ids = { map { $_ => 1 } @$comment_ids };
    
    my $dbh = Bugzilla->dbh;
    if (scalar keys %$comment_ids) {
        my $q = 'SELECT DISTINCT bug_id FROM longdescs WHERE comment_id IN ('
                . join(', ', ('?') x keys(%$comment_ids)) . ')';
        my $comment_bug_ids = $dbh->selectcol_arrayref($q);
        # Set this to 2 so that we can distinguish the two cases
        # (bug wanted vs. comment wanted) later.
        $bug_ids->{$_} ||= 2 foreach (@$comment_bug_ids);
    }

    my $include_fields = $params->{include_fields};
    $include_fields ||= [qw(id number text author time)];
    my %inc = map { $_ => 1 } @$include_fields;

    my $start = $params->{changed_since};
    if (defined $start) {
        # Isn't there a wrapper for this, that the XMLRPC stuff
        # just does automatically???
        $start =~ /^(\d{4})(\d{2})(\d{2})T(\d{2}):(\d{2}):(\d{2})$/
            || ThrowUserError('illegal_date', 
               { date => $start, format => 'YYYYMMDDThh:mm:ss' });
        $start = "$1.$2.$3 $4:$5:$6";
    }

    my %ret;
    foreach my $bug_id (keys %$bug_ids) {
        ValidateBugID($bug_id);

        # Time has to be filtered here, because theres no other way
        # to get the comment count
        my $comments = Bugzilla::Bug::GetComments($bug_id, 'oldest_to_newest');

        my $cnt = -1;

        foreach my $comment (@$comments) {
            $cnt++;

            # Filter to only show wanted comments
            # ie bug wanted or comment wanted
            next unless $bug_ids->{$bug_id} == 1 ||
                $comment_ids->{$comment->{id}};

            # Filter for start time
            next unless $comment->{'time'} ge $start;

            # Filter for comment visibility
            # XXX - This should be filtered somewhere else
            next if $comment->{'isprivate'} && !Bugzilla->user->is_insider;

            my %c;
            $c{id} = type('int')->value($comment->{id}) if $inc{id};
            $c{number} = type('int')->value($cnt) if $inc{number}; 
            $c{text} = type('string')->value($comment->{body}) if $inc{text};
            $c{author} = type('string')->value($comment->{author}->login)
                if $inc{author};
            $c{'time'} = type('dateTime')->value(
                               $self->datetime_format($comment->{'time'}))
                if $inc{'time'};

            push @{$ret{$bug_id}}, \%c;
        }
    }

    return \%ret;
}


sub count {
    my ($self, $params) = @_;
    
    my $products = $params->{products};
    my $dbh = Bugzilla->dbh;

    ThrowUserError('no_products') unless scalar(@$products);
    
    foreach my $product_name (@$products){
        ThrowUserError('product_access_denied', {product => $product_name})
            unless Bugzilla->user->can_see_product($product_name);
    }
    
    # Convert product names to their IDs.
    my $product_ids =
        $dbh->selectcol_arrayref('SELECT id FROM products
                                  WHERE ' .
                                  $dbh->sql_in('name', [('?') x @$products]),
                                  undef, @$products);

    my $bug_count =
      $dbh->selectrow_array('SELECT COUNT(bug_id)
                               FROM bugs
                               WHERE '.
                               $dbh->sql_in('product_id', $product_ids));
                            
    return {'count' => type('int')->value($bug_count)};
}


1;

__END__

=head1 NAME

Bugzilla::Webservice::Bug - The API for creating, changing, and getting the
details of bugs.

=head1 DESCRIPTION

This part of the Bugzilla API allows you to file a new bug in Bugzilla,
or get information about bugs that have already been filed.

=head1 METHODS

See L<Bugzilla::WebService> for a description of B<STABLE>, B<UNSTABLE>,
and B<EXPERIMENTAL>.

=head2 Utility Functions

=over

=item C<legal_values> B<EXPERIMENTAL>

=over

=item B<Description>

Tells you what values are allowed for a particular field.

=item B<Params>

=over

=item C<field> - The name of the field you want information about.
This should be the same as the name you would use in L</create>, below.

=item C<product_id> - If you're picking a product-specific field, you have
to specify the id of the product you want the values for.

=back

=item B<Returns> 

C<values> - An array of strings: the legal values for this field.
The values will be sorted as they normally would be in Bugzilla.

=item B<Errors>

=over

=item 106 (Invalid Product)

You were required to specify a product, and either you didn't, or you
specified an invalid product (or a product that you can't access).

=item 108 (Invalid Field Name)

You specified a field that doesn't exist or isn't a drop-down field.

=back

=back


=back

=head2 Bug Creation and Modification

=over

=item C<get> B<EXPERIMENTAL>

=over

=item B<Description>

Gets information about particular bugs in the database.

Note: Can also be called as "get_bugs" for compatibilty with Bugzilla 3.0 API.

=item B<Params>

=over

=item C<ids>

An array of numbers and strings.

If an element in the array is entirely numeric, it represents a bug_id
from the Bugzilla database to fetch. If it contains any non-numeric 
characters, it is considered to be a bug alias instead, and the bug with 
that alias will be loaded. 

Note that it's possible for aliases to be disabled in Bugzilla, in which
case you will be told that you have specified an invalid bug_id if you
try to specify an alias. (It will be error 100.)

=item C<products>

An array of product names. All bugs in the specified products are returned.

If both C<ids> and C<products> are present, the union of the two are returned.

=item C<changed_since>

A C<dateTime> field. Only bugs changed on or after this time are returned.
Newly-filed bugs are considered "changed" at the time they were filed, and
will be included in the returned results.

=item C<limit>

Return at most this many results.

=item C<offset>

Starts returning bugs from this point in the list. Requires C<limit> to be
set.

Note that while using C<limit> and C<offset> to handle a large list in
chunks is possible, it is impossible to guarantee that no changes are
being made between runs (eg changing the product of a bug), so particular
bugs might be skipped or returned multiple times.

=back

=item B<Returns>

A hash containing a single element, C<bugs>. This is an array of hashes. 
Each hash contains the following items:

=over

=item id

C<int> The numeric bug_id of this bug.

=item alias

C<string> The alias of this bug. If there is no alias or aliases are 
disabled in this Bugzilla, this will be an empty string.

=item summary

C<string> The summary of this bug.

=item creation_time

C<dateTime> When the bug was created.

=item last_change_time

C<dateTime> When the bug was last changed.

=item internals B<UNSTABLE>

A hash. The internals of a L<Bugzilla::Bug> object. This is extremely
unstable, and you should only rely on this if you absolutely have to. The
structure of the hash may even change between point releases of Bugzilla.

=back

=item B<Errors>

=over

=item 100 (Invalid Bug Alias)

If you specified an alias and either: (a) the Bugzilla you're querying
doesn't support aliases or (b) there is no bug with that alias.

=item 101 (Invalid Bug ID)

The bug_id you specified doesn't exist in the database.

=item 102 (Access Denied)

You do not have access to the bug_id you specified.

=item 10001 (Invalid Offset)

You specified an C<offset> that wasn't numeric.

=back

=back



=item C<create> B<EXPERIMENTAL>

=over

=item B<Description>

This allows you to create a new bug in Bugzilla. If you specify any
invalid fields, they will be ignored. If you specify any fields you
are not allowed to set, they will just be set to their defaults or ignored.

You cannot currently set all the items here that you can set on enter_bug.cgi.

The WebService interface may allow you to set things other than those listed
here, but realize that anything undocumented is B<UNSTABLE> and will very
likely change in the future.

=item B<Params>

Some params must be set, or an error will be thrown. These params are
marked B<Required>. 

Some parameters can have defaults set in Bugzilla, by the administrator.
If these parameters have defaults set, you can omit them. These parameters
are marked B<Defaulted>.

Clients that want to be able to interact uniformly with multiple
Bugzillas should always set both the params marked B<Required> and those 
marked B<Defaulted>, because some Bugzillas may not have defaults set
for B<Defaulted> parameters, and then this method will throw an error
if you don't specify them.

The descriptions of the parameters below are what they mean when Bugzilla is
being used to track software bugs. They may have other meanings in some
installations.

=over

=item C<product> (string) B<Required> - The name of the product the bug
is being filed against.

=item C<component> (string) B<Required> - The name of a component in the
product above.

=item C<summary> (string) B<Required> - A brief description of the bug being
filed.

=item C<version> (string) B<Required> - A version of the product above;
the version the bug was found in.

=item C<description> (string) B<Defaulted> - The initial description for 
this bug. Some Bugzilla installations require this to not be blank.

=item C<op_sys> (string) B<Defaulted> - The operating system the bug was
discovered on.

=item C<platform> (string) B<Defaulted> - What type of hardware the bug was
experienced on.

=item C<priority> (string) B<Defaulted> - What order the bug will be fixed
in by the developer, compared to the developer's other bugs.

=item C<severity> (string) B<Defaulted> - How severe the bug is.

=item C<alias> (string) - A brief alias for the bug that can be used 
instead of a bug number when accessing this bug. Must be unique in
all of this Bugzilla.

=item C<assigned_to> (username) - A user to assign this bug to, if you
don't want it to be assigned to the component owner.

=item C<cc> (array) - An array of usernames to CC on this bug.

=item C<qa_contact> (username) - If this installation has QA Contacts
enabled, you can set the QA Contact here if you don't want to use
the component's default QA Contact.

=item C<status> (string) - The status that this bug should start out as.
Note that only certain statuses can be set on bug creation.

=item C<target_milestone> (string) - A valid target milestone for this
product.

=back

In addition to the above parameters, if your installation has any custom
fields, you can set them just by passing in the name of the field and
its value as a string.

=item B<Returns>

A hash with one element, C<id>. This is the id of the newly-filed bug.

=item B<Errors>

=over

=item 51 (Invalid Object)

The component you specified is not valid for this Product.

=item 103 (Invalid Alias)

The alias you specified is invalid for some reason. See the error message
for more details.

=item 104 (Invalid Field)

One of the drop-down fields has an invalid value, or a value entered in a
text field is too long. The error message will have more detail.

=item 105 (Invalid Component)

You didn't specify a component.

=item 106 (Invalid Product)

Either you didn't specify a product, this product doesn't exist, or
you don't have permission to enter bugs in this product.

=item 107 (Invalid Summary)

You didn't specify a summary for the bug.

=item 504 (Invalid User)

Either the QA Contact, Assignee, or CC lists have some invalid user
in them. The error message will have more details.

=back

=item B<History>

=over

=item Before B<3.0.4>, parameters marked as B<Defaulted> were actually
B<Required>, due to a bug in Bugzilla.

=back

=back

=item C<add_comment> B<EXPERIMENTAL>

=over

=item B<Description>

This allows you to add a comment to a bug in Bugzilla.

=item B<Params>

=over

=item C<id> (int) B<Required> - The id or alias of the bug to append a 
comment to.

=item C<comment> (string) B<Required> - The comment to append to the bug.

=item C<private> (boolean) - If set to true, the comment is private, otherwise
it is assumed to be public.

=item C<work_time> (double) - Adds this many hours to the "Hours Worked"
on the bug. If you are not in the time tracking group, this value will
be ignored.


=back

=item B<Returns>

A hash with one element, C<comment_id>. This is the id of the newly-added comment.

=item B<Errors>

=over

=item 100 (Invalid Bug Alias) 

If you specified an alias and either: (a) the Bugzilla you're querying
doesn't support aliases or (b) there is no bug with that alias.

=item 101 (Invalid Bug ID)

The id you specified doesn't exist in the database.

=item 108 (Bug Edit Denied)

You did not have the necessary rights to edit the bug.

=back

=back

=item C<comments> B<UNSTABLE>

=over

=item B<Description>

This allows you to get comments for a set of bugs.

=item B<Params>

=over

=item C<bug_ids> (array of ints) - An array of bugs IDs to return
comments for. At least one of C<bug_ids> or C<comment_ids>
is B<required>.

=item C<comment_ids> (array of ints) - An array of integer comment_ids 
to return. At least one of C<bug_ids> or C<comment_ids>
is B<required>.

=item C<include_fields> (array of strings) - An array of field names to be
included in the returned hash. If this parameter is not present, all fields
are returned. If it is present, then only those fields will be returned. For
forwards/backwards compatibility, it is not an error to specify additional
fields.

=item C<changed_since> (dateTime) - Only return comments that are newer
than the specified date.

=back

=item B<Returns>

A single item, C<bugs>, which is a hash, where the keys are C<bug_id>s and the
values are an array of comments. Each comment is a hash, containing:

=over

=item id

C<int> The globally unique ID for the comment.

=item number

C<int> The comment number for the comment. (You could link to this comment
using C<#c8>, for example, if this was comment number 8 on the bug).

=item text

C<string> The actual text of the comment.

=item author

C<string> The login name of the comment's author.

=item time

C<dateTime> The time (in Bugzilla's timezone) that the comment was added.

=back

=item B<Errors>

Same as L</get>.

=back


=item C<count> B<UNSTABLE>

=over

=item B<Description>

Returns the bug count for a list of products.

=item B<Params>

=over

=item C<products> (array) B<Required> - An array of product ID's.

=back

=item B<Returns>

A hash with one element, C<count>. This is the bug count for the list of products.

=item B<Errors>

=over

=item 106 (Invalid Product)

Either you didn't specify a product, one or more products don't exist, or
you don't have permission to access bugs for one or more of the products.

=back

=back

=back
