# -*- Mode: perl; indent-tabs-mode: nil -*-
#
# The contents of this file are subject to the Mozilla Public
# License Version 1.1 (the "License"); you may not use this file
# except in compliance with the License. You may obtain a copy of
# the License at http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS
# IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
# implied. See the License for the specific language governing
# rights and limitations under the License.
#
# The Original Code is the Bugzilla Launchpad Plugin.
#
# The Initial Developer of the Original Code is Canonical Ltd.
# Portions created by Canonical are Copyright (C) 2008 Canonical Ltd.
# All Rights Reserved.
#
# Contributor(s): Bradley Baetz <bbaetz@acm.org>
#                 Elliotte Martin <emartin@everythingsolved.com>

package extensions::launchpad::lib::WSLaunchpad;
use strict;
use warnings;

use base qw(Bugzilla::WebService);

import SOAP::Data qw(type);

use Bugzilla::Bug;
use Bugzilla::Constants;
use Bugzilla::Error;
use Bugzilla::Util;
use Bugzilla::WebService::Bugzilla;

use Date::Format;
use LWP::UserAgent;
use Time::Zone;

use constant LAUNCHPAD_TIMEOUT => 10; # Seconds before token-check timeout.
use constant LAUNCHPAD_URL => 'https://launchpad.net';

use constant LOGIN_EXEMPT => {
        login => 1,
};

sub set_link {
    my ($self, $params) = @_;
    
    my $bug_id = $params->{id};
    my $launchpad_id = $params->{launchpad_id};

    #The user must login in order add a launchpad_id
    Bugzilla->login(LOGIN_REQUIRED);

    # Check parameters
    defined $bug_id  || ThrowCodeError('missing_bug_id');

    ValidateBugID($bug_id);

    my $bug = new Bugzilla::Bug($bug_id);

    Bugzilla->user->can_edit_product($bug->product_id)
        || ThrowUserError("product_edit_denied", {product => $bug->product});

    my $dbh = Bugzilla->dbh;
    
    my $old_launchpad_id = $dbh->selectrow_array(
            "SELECT launchpad_id
               FROM bugs WHERE bug_id = ?", undef, $bug_id);

    my $privs;
    $bug->check_can_change_field('launchpad_id', 
                                 $old_launchpad_id, $launchpad_id, \$privs)
            || ThrowUserError('illegal_change', 
                              { field => 'launchpad_id', privs => $privs });

    if (defined ($launchpad_id)) {
        if (!defined ($old_launchpad_id) ||
             $launchpad_id != $old_launchpad_id)
        {
               $dbh->do("UPDATE bugs SET launchpad_id = ?
                         WHERE bug_id = ?", undef, $launchpad_id, $bug_id);
        }
    }
    elsif (defined ($old_launchpad_id)){
            $dbh->do("UPDATE bugs SET launchpad_id = NULL
                      WHERE bug_id = ?", undef, $bug_id);
    }

    return {'launchpad_id' => type('int')->value($old_launchpad_id)};
}

sub login {
    my ($self, $params) = @_;

    my $token = $params->{token};
    $token || ThrowCodeError('param_required', { param => 'token' });

    # Taken from Bugzilla::Update
    my $ua = LWP::UserAgent->new();
    $ua->timeout(LAUNCHPAD_TIMEOUT);
    $ua->protocols_allowed(['https']);
    # If the URL of the proxy is given, use it, else get this information
    # from the environment variable.
    my $proxy_url = Bugzilla->params->{'proxy_url'};
    if ($proxy_url) {
        $ua->proxy(['http', 'https'], $proxy_url);
    } else {
        $ua->env_proxy;
    }

    my $resp = $ua->post(LAUNCHPAD_URL . 'token/' . url_quote($token)
                         . '/+bugtracker-handshake');
    
    if ($resp->code() != 200) {
        ThrowUserError('launchpad_invalid_token', {
            token => $token,
            code => $resp->code,
            content => $resp->content,
        });
    }

    my $auth = Bugzilla::Auth->new({Login => 'Launchpad'});
    my $user = $auth->login(LOGIN_REQUIRED);

    return { user_id => type('int')->value($user->id) };
}

sub plugin_version(){	
    my $plugins = Bugzilla::WebService::Bugzilla::extensions();
    return {'version' => $plugins->{'extensions'}->{'launchpad'}->{'version'}};
}

sub time {
    my ($self) = @_;
    my $dbh = Bugzilla->dbh;

    my $db_time = $dbh->selectrow_array('SELECT LOCALTIMESTAMP(0)');
    return {
        local_time => type('dateTime')->value($self->datetime_format($db_time)),
        utc_time   => type('dateTime')->value(
                          time2str('%Y%m%dT%H:%M:%S', time, 'UTC')),
        tz_name => type('string')->value(uc tz_name()),
    };
}

1;

__END__

=head1 NAME

extensions::launchpad::lib::WSLaunchpad - Webservice functions for the
Launchpad extention

=head1 DESCRIPTION

This package contains various webservices calls for the Launchpad plugin

=head1 CALLS

There are a number of XML-RPC calls.

=over

=item C<login> B<EXPERIMENTAL>

=over

=item B<Description>

Allows launchpad to log in. The launchpad server URL and Bugzilla username
must be configured in the preferences.

This routine takes a token provided by the remote server. Bugzilla does a
POST to https://launchpad.net/token/C<token>/+bugtracker-handshake using 
the provided token.

If that POST returns a 200 response, then the token is accepted, and Bugzilla
returns the standard authentication cookie headers. Otherwise, the request
fails.

The returned headers are then valid in any further request to serve as
authentication. Note that the Bugzilla server tends to expire these logins
after 30 days, but they could expire earlier than that if the admin manually
logs out all users. (It is impossible to know whether or not your cookie
is still valid without using it.)

=item B<Params>

=over

=item C<token> (string) B<Required> - The token to use.

=back

=item B<Returns>

=over

=item C<user_id> (integer) - The user_id of the configured launchpad user.

=back

=item B<Errors>

=over

=item 10300 (Invalid Token)

The token you provided was not valid (the login failed).

=item -10301 (Plugin Not Configured)

The user hasn't run F<checksetup.pl> since installing the Launchpad
plugin.

=back

=back

=item C<set_link> B<EXPERIMENTAL>

=over

=item B<Description>

Takes a Launchpad bug ID and a Bugzilla Bug ID and sets a link between them.
If the Launchpad bug ID is not specified, any existing link will be deleted.

=item B<Params>

=over

=item C<id> (int) B<Required> - The id or alias of the Bugzilla bug.

=item C<launchpad_id> (int) - The Launchpad bug ID.

=back

=item B<Returns>

C<launchpad_id> (int) The previously-existing Launchpad ID, if there was one.

=item B<Errors>

=over

=item 100 (Invalid Bug Alias)

If you specified an alias and either: (a) the Bugzilla you're querying
doesn't support aliases or (b) there is no bug with that alias.

=item 101 (Invalid Bug ID)

The id you specified doesn't exist in the database.

=item 108 (Bug Edit Denied)

You did not have the necessary rights to edit the bug.

=back

=back

=item C<time> B<EXPERIMENTAL>

Returns three elements:

=over

=item local_time (dateTime)

The current time in the local timezone, according to the Bugzilla
database server.

=item utc_time (dateTime)

The current time in UTC, according to the Bugzilla web server.

=item tz_name (string)

Current time zone

=back

=item C<plugin_version> B<EXPERIMENTAL>

=over

=item B<Description>

Returns the Launchpad plug-in version number.

=item B<Returns>

A hash with a single item, C<version>, that is the version as a
string.

=back

=back
